/*
 * Created on 24-nov-2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package org.herac.tuxguitar.play.models.midiplayer;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiChannel;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Receiver;
import javax.sound.midi.Sequencer;
import javax.sound.midi.Soundbank;
import javax.sound.midi.Synthesizer;
import javax.sound.midi.Transmitter;

import org.herac.tuxguitar.play.models.Player;
import org.herac.tuxguitar.song.managers.SongManager;
import org.herac.tuxguitar.song.models.InstrumentString;
import org.herac.tuxguitar.song.models.Note;
import org.herac.tuxguitar.song.models.SongChannel;
import org.herac.tuxguitar.song.models.SongTrack;

/**
 * @author julian
 * 
 * TODO To change the template for this generated type comment go to Window - Preferences - Java - Code Style - Code Templates
 */
public class SongPlayer implements Player {
    private static int DELAY = 100;
    
    private SongManager songManager;

    private Sequencer sequencer;

    private Synthesizer synthesizer;

    private Soundbank soundbank;

    private MeasureStartMetaEventListener controller;

    private boolean running;

    private boolean changeTickPosition;

    private long tickPosition;

    public SongPlayer(SongManager songManager) {
        this.songManager = songManager;
        this.controller = new MeasureStartMetaEventListener();
        this.init();
        this.reset();
    }

    /**
     * Retorna el Sequenciador
     * TODO: si es java 1.5 inicializar desconectado: MidiSystem.getSequencer(false);
     */
    private Sequencer getSequencer() {
        try {
            if (this.sequencer == null) {
                this.sequencer = MidiSystem.getSequencer(false);
                this.sequencer.addMetaEventListener(this.controller);
            }
            if (!this.sequencer.isOpen()) {
                this.sequencer.open();
            }
            return this.sequencer;
        } catch (MidiUnavailableException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * Retorna el Sintetizador
     */
    public Synthesizer getSynthesizer() {
        if (this.synthesizer == null) {
            Sequencer sequencer = getSequencer();
            if (sequencer instanceof Synthesizer) {
                this.synthesizer = (Synthesizer) sequencer;
            } else {
                try {
                    this.synthesizer = MidiSystem.getSynthesizer();
                    this.synthesizer.open();
                    Receiver receiver = this.synthesizer.getReceiver();
                    Transmitter transmitter = sequencer.getTransmitter();
                    transmitter.setReceiver(receiver);
                } catch (MidiUnavailableException e) {
                    e.printStackTrace();
                }
            }
        }
        return this.synthesizer;
    }
    
    /**
     * Retorna el Soundbank por defecto
     */
    public Soundbank getDefaultSoundbank() {
        if (this.soundbank == null) {
            Synthesizer synthesizer = getSynthesizer();
            this.soundbank = synthesizer.getDefaultSoundbank();
        }
        return this.soundbank;
    }
    
    /**
     * Inicia el Secuenciador y Sintetizador
     */

    public void init() {
        getSynthesizer();
        getSequencer();
    }

    /**
     * Resetea los valores
     */
    public void reset() {
        this.stop();
        this.tickPosition = 1000;
        this.setChangeTickPosition(false);
        this.controller.reset();
    }

    /**
     * Cierra el Secuenciador y Sintetizador
     */
    public void close() {
        Synthesizer synthesizer = getSynthesizer();
        if (synthesizer != null) {
            synthesizer.close();
        }        
        Sequencer sequencer = getSequencer();
        if (sequencer != null) {
            sequencer.close();
        }
        setRunning(false);
    }

    /**
     * Para la reproduccion
     */
    public void stop() {
        Sequencer sequencer = getSequencer();
        if (sequencer != null && sequencer.isRunning()) {
            sequencer.stop();
        }
        this.setRunning(false);
    }

    /**
     * Inicia la reproduccion
     */
    public void play() {
        this.stop();
        this.setRunning(true);        
        new Thread(new Runnable() {
            public void run() {
                try {                    
                    addSecuence();                    
                    getSequencer().start();
                    getSequencer().setTickPosition(tickPosition - DELAY);
                    setChangeTickPosition(false);
                    updateChannels();
                    //updateVolumes();
                    //updateBalances();
                    while (getSequencer().isRunning()) {
                        if (isChangeTickPosition()) {
                            getSequencer().setTickPosition(tickPosition - DELAY);
                            setChangeTickPosition(false);
                            updateChannels();
                            //updateVolumes();
                        }
                        Thread.sleep(10);
                    }

                } catch (InterruptedException e) {
                    e.printStackTrace();
                } catch (InvalidMidiDataException e) {
                    close();
                    init();
                    e.printStackTrace();
                } finally {
                    reset();
                }

            }
        }).start();

    }

    /**
     * Indica la posicion del secuenciador
     */
    public void setTickPosition(long position) {
        this.tickPosition = position;
        this.setChangeTickPosition(true);
    }

    /**
     * Asigna el valor a running
     */
    public void setRunning(boolean running) {
        this.running = running;
    }

    /**
     * Retorna True si esta reproduciendo
     */
    public boolean isRunning() {
        return this.running;
    }

    /**
     * Retorna True si hay cambios en la posicion
     */
    private boolean isChangeTickPosition() {
        return changeTickPosition;
    }

    /**
     * Asigna los cambios de la posicion
     */
    private void setChangeTickPosition(boolean changeTickPosition) {
        this.changeTickPosition = changeTickPosition;
    }

    /**
     * Retorna el tick de la nota que esta reproduciendo
     */
    public long getTickPosition() {
        if(this.sequencer.getTickPosition() >= this.tickPosition + DELAY){
            return this.sequencer.getTickPosition() - this.controller.getStartMove();
        }
        return 0;
    }

    /**
     * Agrega la Secuencia
     */
    public void addSecuence() throws InvalidMidiDataException {
        SongSequence songSequence = new SongSequence(songManager);
        songSequence.createSongSecuence();
        getSequencer().setSequence(songSequence.getSongSecuence());
    }
    
    
    public void loadSoundbank(File file){    
		try {
			Soundbank soundbank = MidiSystem.getSoundbank(file);

			if (soundbank != null){
				if(getSynthesizer().isSoundbankSupported(soundbank)){
					getSynthesizer().loadAllInstruments(soundbank);
					this.soundbank = soundbank;
				}
			}    	
		} catch (InvalidMidiDataException e) {			
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
    }
    
    public void updateChannels(){
        Iterator it = this.songManager.getSong().getTracks().iterator();
        while(it.hasNext()){
            SongTrack track = (SongTrack)it.next();
            updateChannel(track.getChannel());            
        }
    }
    
    public void updateChannel(SongChannel channel) {
    	double gain = (double)(this.songManager.getSong().getVolume() / 10.0);
    	int volume = (int)(gain * channel.getVolume());      	
    	int balance = channel.getBalance();   	
    	boolean solo = channel.isSolo();
    	boolean mute = channel.isMute();
    	
        Synthesizer synthesizer = getSynthesizer();
        MidiChannel[] channels = synthesizer.getChannels();   
        
        channels[channel.getChannel()].controlChange(DefinedControllers.VOLUME,volume);
        channels[channel.getChannel()].controlChange(DefinedControllers.BALANCE,balance);
        channels[channel.getChannel()].setSolo(solo);
        channels[channel.getChannel()].setMute(mute);
        if(channel.getChannel() != channel.getEffectChannel()){
        	channels[channel.getEffectChannel()].controlChange(DefinedControllers.VOLUME,volume);
        	channels[channel.getEffectChannel()].controlChange(DefinedControllers.BALANCE,balance);
        	channels[channel.getEffectChannel()].setSolo(solo);
        	channels[channel.getEffectChannel()].setMute(mute);
        }        
    }

    
    
    
    
    public void playBeat(final SongTrack track,final List notes){    			
		final MidiChannel[] channels = getSynthesizer().getChannels(); 
		Iterator it = notes.iterator();
		while(it.hasNext()){
			Note note = (Note)it.next();							    	
			final int noteValue = (note.getValue() + ((InstrumentString)track.getStrings().get(note.getString() - 1)).getValue());
			final int channel = (note.getEffect().hasEffects())?track.getChannel().getEffectChannel():track.getChannel().getChannel();
			new Thread(new Runnable() {		
				public void run() {
					try {								
						channels[channel].controlChange(DefinedControllers.VOLUME,127);
						channels[channel].controlChange(DefinedControllers.BALANCE,64);					
						channels[channel].programChange(track.getChannel().getInstrument());					
						channels[channel].noteOn(noteValue,64);		
									
						Thread.sleep(750);
									
						channels[channel].noteOff(noteValue,32);		
					} catch (InterruptedException e) {					
						e.printStackTrace();
					}
				}		
			}).start();
		}					

    }
    
}